#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import  # needed for zmq import
import zmq

from . import ServerTransport, ClientTransport


class ZmqServerTransport(ServerTransport):
    """Server transport based on a :py:const:`zmq.ROUTER` socket.

    :param socket: A :py:const:`zmq.ROUTER` socket instance, bound to an
                   endpoint.
    """

    def __init__(self, socket):
        self.socket = socket

    def receive_message(self):
        msg = self.socket.recv_multipart()
        return msg[:-1], msg[-1]

    def send_reply(self, context, reply):
        self.socket.send_multipart(context + [reply])

    @classmethod
    def create(cls, zmq_context, endpoint):
        """Create new server transport.

        Instead of creating the socket yourself, you can call this function and
        merely pass the :py:class:`zmq.core.context.Context` instance.

        By passing a context imported from :py:mod:`zmq.green`, you can use
        green (gevent) 0mq sockets as well.

        :param zmq_context: A 0mq context.
        :param endpoint: The endpoint clients will connect to.
        """
        socket = zmq_context.socket(zmq.ROUTER)
        socket.bind(endpoint)
        return cls(socket)


class ZmqClientTransport(ClientTransport):
    """Client transport based on a :py:const:`zmq.REQ` socket.

    :param socket: A :py:const:`zmq.REQ` socket instance, connected to the
                   server socket.
    """

    def __init__(self, socket):
        self.socket = socket

    def send_message(self, message, expect_reply=True):
        self.socket.send(message)

        if expect_reply:
            return self.socket.recv()

    @classmethod
    def create(cls, zmq_context, endpoint):
        """Create new client transport.

        Instead of creating the socket yourself, you can call this function and
        merely pass the :py:class:`zmq.core.context.Context` instance.

        By passing a context imported from :py:mod:`zmq.green`, you can use
        green (gevent) 0mq sockets as well.

        :param zmq_context: A 0mq context.
        :param endpoint: The endpoint the server is bound to.
        """
        socket = zmq_context.socket(zmq.REQ)
        socket.connect(endpoint)
        return cls(socket)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import requests

from . import ClientTransport


class HttpPostClientTransport(ClientTransport):
    """HTTP POST based client transport.

    Requires :py:mod:`requests`. Submits messages to a server using the body of
    an ``HTTP`` ``POST`` request. Replies are taken from the responses body.

    :param str endpoint: The URL to send ``POST`` data to.
    :param callable post_method: allows to replace `requests.post` with another method,
        e.g. the post method of a `requests.Session()` instance.
    :param dict kwargs: Additional parameters for :py:func:`requests.post`.
    """

    def __init__(self, endpoint, post_method=None, **kwargs):
        self.endpoint = endpoint
        self.request_kwargs = kwargs
        if post_method is None:
            self.post = requests.post
        else:
            self.post = post_method

    def send_message(self, message, expect_reply=True):
        if not isinstance(message, bytes):
            raise TypeError('message must by of type bytes')

        r = self.post(self.endpoint, data=message, **self.request_kwargs)

        if expect_reply:
            # Note that this is not strictly conforming to the (JSON RPC) standard since
            # even notifications may, under certain conditions, return an
            # error message which is completely ignored by this implementation.
            return r.content



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import queue

from . import ServerTransport
from geventwebsocket.resource import WebSocketApplication, Resource


class WSServerTransport(ServerTransport):
    """
    Requires :py:mod:`geventwebsocket`.

    Due to the nature of WS, this transport has a few pecularities: It must
    be run in a thread, greenlet or some other form of concurrent execution
    primitive.

    This is due to
    :py:attr:`~tinyrpc.transports.websocket.WSServerTransport.handle` which is
    a :py:class:`geventwebsocket.resource.Resource` that joins a wsgi handler
    for the / and a WebSocket handler for the /ws path. These resource is
    used in combination with a :py:class:`geventwebsocket.server.WebSocketServer`
    that blocks while waiting for a call to
    :py:func:`~tinyrpc.transports.wsgi.WSServerTransport.send_reply`.

    The parameter ``queue_class`` must be used to supply a proper queue class
    for the chosen concurrency mechanism (i.e. when using :py:mod:`gevent`,
    set it to :py:class:`gevent.queue.Queue`).

    :param queue_class: The queue class to use.
    :param wsgi_handler: Can be used to change the standard response to a
        http request to the /
    """

    def __init__(self, queue_class=queue.Queue, wsgi_handler=None):
        self._queue_class = queue_class
        self.messages = queue_class()

        def static_wsgi_app(environ, start_response):
            start_response("200 OK", [("Content-Type", "text/html")])
            return 'Ready for WebSocket connection in /ws'

        self.handle = Resource({
            '/':
            static_wsgi_app if wsgi_handler is None else wsgi_handler,
            '/ws':
            WSApplicationFactory(self.messages, queue_class)
        })

    def receive_message(self):
        return self.messages.get()

    def send_reply(self, context, reply):
        context.put(reply)


class WSApplicationFactory(object):
    '''
    Creates WebSocketApplications with a messages queue and the queue_class
    needed for the communication with the WSServerTransport.
    '''

    def __init__(self, messages, queue_class):
        self.messages = messages
        self._queue_class = queue_class

    def __call__(self, ws):
        '''
        The fake __init__ for the WSApplication
        '''
        app = WSApplication(ws)
        app.messages = self.messages
        app._queue_class = self._queue_class
        return app

    @classmethod
    def protocol(cls):
        return WebSocketApplication.protocol()


class WSApplication(WebSocketApplication):
    '''
    This class is the bridge between the WSServerTransport and the WebSocket
    protocol implemented by
    :py:class:`geventwebsocket.resource.WebSocketApplication`
    '''

    def on_message(self, msg, *args, **kwargs):
        # create new context
        context = self._queue_class()
        self.messages.put((context, msg))
        response = context.get()
        self.ws.send(response, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Protocol definition.

Defines the abstract base classes from which a protocol definition must be constructed.
"""

from .. import exc


class RPCRequest(object):
    """Defines a generic RPC request."""

    def __init__(self):
        self.unique_id = None
        """Correlation ID used to match request and response.
        
        :type: int or str or None
        
        Protocol specific, may or may not be set.
        This value should only be set by :py:func:`~tinyrpc.protocols.RPCProtocol.create_request`.

        When the protocol permits it this ID allows servers to respond to requests out
        of order and allows clients to relate a response to the corresponding request.

        Only supported if the protocol has its
        :py:attr:`~tinyrpc.protocols.RPCProtocol.supports_out_of_order` set to ``True``.
        
        Generated by the client, the server copies it from request to corresponding response.
        """

        self.method = None
        """The name of the RPC function to be called.
        
        :type: str
        
        The :py:attr:`method` attribute uses the name of the function as it is known by the public.
        The :py:class:`~tinyrpc.dispatch.RPCDispatcher` allows the use of public aliases in the
        ``@public`` decorators.
        These are the names used in the :py:attr:`method` attribute.
        """

        self.args = []
        """The positional arguments of the method call.
        
        :type: list
        
        The contents of this list are the positional parameters for the :py:attr:`method` called.
        It is eventually called as ``method(*args)``.
        """

        self.kwargs = {}
        """The keyword arguments of the method call.
        
        :type: dict
        
        The contents of this dict are the keyword parameters for the :py:attr:`method` called.
        It is eventually called as ``method(**kwargs)``.
        """

    def error_respond(self, error):
        """Creates an error response.

        Create a response indicating that the request was parsed correctly,
        but an error has occured trying to fulfill it.

        This is an abstract method that must be overridden in a derived class.

        :param error: An exception or a string describing the error.
        :type error: Exception or str

        :return: A response or ``None`` to indicate that no error should be sent out.
        :rtype: :py:class:`RPCErrorResponse` or None
        """
        raise NotImplementedError()

    def respond(self, result):
        """Create a response.

        Call this to return the result of a successful method invocation.

        This creates and returns an instance of a protocol-specific subclass of
        :py:class:`~tinyrpc.RPCResponse`.

        This is an abstract method that must be overridden in a derived class.

        :param result: Passed on to new response instance.
        :type result: Any type that can be serialized by the protocol.

        :return: A response or ``None`` to indicate this request does not expect a response.
        :rtype: :py:class:`RPCResponse` or None
        """
        raise NotImplementedError()

    def serialize(self):
        """Returns a serialization of the request.

        Converts the request into a bytes object that can be passed to and by the transport layer.

        This is an abstract method that must be overridden in a derived class.

        :return: A bytes object to be passed on to a transport.
        :rtype: bytes
        """
        raise NotImplementedError()


class RPCBatchRequest(list):
    """Multiple requests batched together.

    Protocols that support multiple requests in a single message use this to group them together.
    Note that not all protocols may support batch requests.

    Handling a batch requests is done in any order, responses must be gathered
    in a batch response and be in the same order as their respective requests.

    Any item of a batch request is either an :py:class:`RPCRequest` or an
    :py:class:`~tinyrpc.exc.BadRequestError`, which indicates that there has been
    an error in parsing the request.
    """

    def create_batch_response(self):
        """Creates a response suitable for responding to this request.

        This is an abstract method that must be overridden in a derived class.

        :return: An :py:class:`RPCBatchResponse` or None if no response is expected.
        :rtype: :py:class:`RPCBatchResponse` or None
        """
        raise NotImplementedError()

    def serialize(self):
        """Returns a serialization of the request.

        Converts the request into a bytes object that can be passed to and by the transport layer.

        This is an abstract method that must be overridden in a derived class.

        :return: A bytes object to be passed on to a transport.
        :rtype: bytes
        """
        raise NotImplementedError()


class RPCResponse(object):
    """Defines a generic RPC response.

    Base class for all deriving responses.

    .. py:attribute:: id

        Correlation ID to match request and response

        :type: str or int

    .. py:attribute:: result

        When present this attribute contains the result of the RPC call.
        Otherwise the :py:attr:`error` attribute must be defined.

        :type: Any type that can be serialized by the protocol.

    .. py:attribute:: error

        When present the :py:attr:`result` attribute must be absent.
        Presence of this attribute indicates an error condition.

        :type: :py:class:`~tinyrpc.exc.RPCError`
    """

    def __init__(self):
        self.unique_id = None
        """Correlation ID used to match request and response.

        :type: int or str or None
        """

    def serialize(self):
        """Returns a serialization of the response.

        Converts the response into a bytes object that can be passed to and by the transport layer.

        This is an abstract method that must be overridden in a derived class.

        :return: The serialized encoded response object.
        :rtype: bytes
        """
        raise NotImplementedError()


class RPCErrorResponse(RPCResponse):
    """RPC error reponse class.

    Base class for all deriving responses.

    .. py:attribute:: error

        This attribute contains the fields ``message`` (str) and
        ``code`` (int) where at least ``message`` is required to contain a value.

        :type: dict
    """
    error = None


class RPCBatchResponse(list):
    """Multiple response from a batch request. See
    :py:class:`RPCBatchRequest` on how to handle.

    Items in a batch response need to be
    :py:class:`RPCResponse` instances or None, meaning no reply should
    generated for the request.
    """

    def serialize(self):
        """Returns a serialization of the batch response.

        Converts the response into a bytes object that can be passed to and by the transport layer.

        This is an abstract method that must be overridden in a derived class.

        :return: A bytes object to be passed on to a transport.
        :rtype: bytes
        """
        raise NotImplementedError()


class RPCProtocol(object):
    """Abstract base class for all protocol implementations."""

    supports_out_of_order = False
    """If true, this protocol can receive responses out of order correctly.

    Note that this usually depends on the generation of unique_ids, the
    generation of these may or may not be thread safe, depending on the
    protocol. Ideally, only one instance of RPCProtocol should be used per
    client.
    
    :type: bool
    """

    raises_errors = True
    """If True, this protocol instance will raise an RPCError exception.

    On receipt of an RPCErrorResponse instance an RPCError exception is raised.
    When this flag is False the RPCErrorReponse object is returned to the caller
    which is then responsible for handling the error.
    
    :type: bool
    """

    def create_request(self, method, args=None, kwargs=None, one_way=False):
        """Creates a new :py:class:`RPCRequest` object.

        Called by the client when constructing a request.
        It is up to the implementing protocol whether or not ``args``,
        ``kwargs``, one of these, both at once or none of them are supported.

        :param str method: The method name to invoke.
        :param list args: The positional arguments to call the method with.
        :param dict kwargs: The keyword arguments to call the method with.
        :param bool one_way: The request is an update, i.e. it does not expect a reply.
        :return: A new request instance.
        :rtype: :py:class:`RPCRequest`
        """
        raise NotImplementedError()

    def parse_request(self, data):
        """Deserializes and validates a request.

        Called by the server to reconstruct the serialized :py:class:`RPCRequest`.

        :param bytes data: The data stream received by the transport layer containing the
            serialized request.
        :return: A reconstructed request.
        :rtype: :py:class:`RPCRequest`
        """
        raise NotImplementedError()

    def parse_reply(self, data):
        """Deserializes and validates a response.

        Called by the client to reconstruct the serialized :py:class:`RPCResponse`.

        :param bytes data: The data stream received by the transport layer containing the
            serialized response.
        :return: A reconstructed response.
        :rtype: :py:class:`RPCResponse`
        """
        raise NotImplementedError()

    def raise_error(self, error):
        """Raises the exception in the client.

        Called by the client to convert the :py:class:`RPCErrorResponse` into an Exception
        and raise or return it depending on the :py:attr:`raises_errors` attribute.

        :param error: The error response received from the server.
        :type error: :py:class:`RPCResponse`
        :rtype: :py:exc:`~tinyrpc.exc.RPCError` when :py:attr:`raises_errors` is False.
        :raises: :py:exc:`~tinyrpc.exc.RPCError` when :py:attr:`raises_errors` is True.
        """
        ex = exc.RPCError('Error calling remote procedure: %s' % error.error['message'])
        if self.raises_errors:
            raise ex
        return ex


class RPCBatchProtocol(RPCProtocol):
    def create_batch_request(self, requests=None):
        """Create a new :py:class:`RPCBatchRequest` object.

        Called by the client when constructing a request.

        :param requests: A list of requests.
        :type requests: :py:class:`list` of :py:class:`RPCRequest`
        :return: A new request instance.
        :rtype: :py:class:`RPCBatchRequest`
        """
        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import geventwebsocket as websocket

from . import ClientTransport


class HttpWebSocketClientTransport(ClientTransport):
    """HTTP WebSocket based client transport.

    Requires :py:mod:`websocket-python`. Submits messages to a server using the body of
    an ``HTTP`` ``WebSocket`` message. Replies are taken from the response of the websocket.

    The connection is establish on the ``__init__`` because the protocol is connection oriented,
    you need to close the connection calling the close method.

    :param str endpoint: The URL to connect the websocket.
    :param dict kwargs: Additional parameters for :py:func:`websocket.send`.
    """

    def __init__(self, endpoint, **kwargs):
        self.endpoint = endpoint
        self.request_kwargs = kwargs
        self.ws = websocket.create_connection(self.endpoint, **kwargs)

    def send_message(self, message, expect_reply=True):
        if not isinstance(message, bytes):
            raise TypeError('message must by of type bytes')
        self.ws.send(message)
        r = self.ws.recv()
        if expect_reply:
            return r

    def close(self):
        """Terminate the connection.

        Since WebSocket maintains an open connection over multiple calls
        it must be closed explicitly.
        """
        if self.ws is not None:
            self.ws.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""JSON RPC 2.0 Protocol implementation.

This module can use the jsonext_ package to make it easier to convert to JSON.
In order to use jsonext import it before importing tinyrpc.
Tinyrpc will detect the presence of jsonext and use it automatically.

.. _jsonext: https://pypi.org/project/jsonext

"""

import json
import sys

from .. import (
    RPCBatchProtocol, RPCRequest, RPCResponse, RPCErrorResponse,
    InvalidRequestError, MethodNotFoundError, InvalidReplyError, RPCError,
    RPCBatchRequest, RPCBatchResponse, InvalidParamsError
)

if 'jsonext' in sys.modules:
    # jsonext was imported before this file, assume the intent is that
    # it is used in place of the regular json encoder.
    import jsonext

    json_dumps = jsonext.dumps
else:
    json_dumps = json.dumps


class FixedErrorMessageMixin(object):
    """Combines JSON RPC exceptions with the generic RPC exceptions.

    Constructs the exception using the provided parameters as well as
    properties of the JSON RPC Exception.

    JSON RPC exceptions declare two attributes:

    .. py:attribute:: jsonrpc_error_code

        This is an error code conforming to the JSON RPC `error codes`_ convention.

        :type: :py:class:`int`

    .. py:attribute:: message

        This is a textual representation of the error code.

        :type: :py:class:`str`

    :param list args: Positional arguments for the constructor.
        When present it overrules the :py:attr:`message` attribute.
    :param dict kwargs: Keyword arguments for the constructor.
        If the ``data`` parameter is found in ``kwargs`` its contents are
        used as the *data* property of the JSON RPC Error object.

    :py:class:`FixedErrorMessageMixin` is the basis for adding your own
    exceptions to the predefined ones.
    Here is a version of the reverse string example that dislikes palindromes:

    .. code-block:: python

        class PalindromeError(FixedErrorMessageMixin, Exception)
            jsonrpc_error_code = 99
            message = "Ah, that's cheating"

        @public
        def reverse_string(s):
            r = s[::-1]
            if r == s:
                raise PalindromeError(data=s)
            return r

    >>> client.reverse('rotator')

    Will return an error object to the client looking like:

    .. code-block:: json

        {
            "jsonrpc": "2.0",
            "id": 1,
            "error": {
                "code": 99,
                "message": "Ah, that's cheating",
                "data": "rotator"
            }
        }

    .. _error codes: https://www.jsonrpc.org/specification#error_object
    """

    def __init__(self, *args, **kwargs):
        if not args:
            args = [self.message]
        if 'data' in kwargs:
            self.data = kwargs.pop('data')
        super(FixedErrorMessageMixin, self).__init__(*args, **kwargs)

    def error_respond(self):
        """Converts the error to an error response object.

        :return: An error response object ready to be serialized and sent to the client.
        :rtype: :py:class:`JSONRPCErrorResponse`
        """
        response = JSONRPCErrorResponse()

        response.error = self.message
        response.unique_id = None
        response._jsonrpc_error_code = self.jsonrpc_error_code
        if hasattr(self, 'data'):
            response.data = self.data
        return response


class JSONRPCParseError(FixedErrorMessageMixin, InvalidRequestError):
    """The request cannot be decoded or is malformed."""
    jsonrpc_error_code = -32700
    message = 'Parse error'


class JSONRPCInvalidRequestError(FixedErrorMessageMixin, InvalidRequestError):
    """The request contents are not valid for JSON RPC 2.0"""
    jsonrpc_error_code = -32600
    message = 'Invalid Request'


class JSONRPCMethodNotFoundError(FixedErrorMessageMixin, MethodNotFoundError):
    """The requested method name is not found in the registry."""
    jsonrpc_error_code = -32601
    message = 'Method not found'


class JSONRPCInvalidParamsError(FixedErrorMessageMixin, InvalidRequestError):
    """The provided parameters are not appropriate for the function called."""
    jsonrpc_error_code = -32602
    message = 'Invalid params'


class JSONRPCInternalError(FixedErrorMessageMixin, InvalidRequestError):
    """Unspecified error, not in the called function."""
    jsonrpc_error_code = -32603
    message = 'Internal error'


class JSONRPCServerError(FixedErrorMessageMixin, InvalidRequestError):
    """Unspecified error, this message originates from the called function."""
    jsonrpc_error_code = -32000
    message = ''


class JSONRPCError(FixedErrorMessageMixin, RPCError):
    """Reconstructs (to some extend) the server-side exception.

    The client creates this exception by providing it with the ``error``
    attribute of the JSON error response object returned by the server.

    :param dict error: This dict contains the error specification:

        * code (int): the numeric error code.
        * message (str): the error description.
        * data (any): if present, the data attribute of the error
    """

    def __init__(self, error):
        if isinstance(error, JSONRPCErrorResponse):
            super(JSONRPCError, self).__init__(error.error)
            self._jsonrpc_error_code = error._jsonrpc_error_code
            if hasattr(error, 'data'):
                self.data = error.data
        else:
            super(JSONRPCError, self).__init__()
            self.message = error['message']
            self._jsonrpc_error_code = error['code']
            if 'data' in error:
                self.data = error['data']


class JSONRPCSuccessResponse(RPCResponse):
    """Collects the attributes of a succesful response message.

    Contains the fields of a normal (i.e. a non-error) response message.

    .. py:attribute:: id

        Correlation ID to match request and response.
        A JSON RPC response *must* have a defined matching id attribute.
        ``None`` is not a valid value for a successful response.

        :type: str or int

    .. py:attribute:: result

        Contains the result of the RPC call.

        :type: Any type that can be serialized by the protocol.
    """

    def _to_dict(self):
        return {
            'jsonrpc': JSONRPCProtocol.JSON_RPC_VERSION,
            'id': self.unique_id,
            'result': self.result
        }

    def serialize(self):
        """Returns a serialization of the response.

        Converts the response into a bytes object that can be passed to and by the transport layer.

        :return: The serialized encoded response object.
        :rtype: bytes
        """
        return json_dumps(self._to_dict()).encode()


class JSONRPCErrorResponse(RPCErrorResponse):
    """Collects the attributes of an error response message.

    Contains the fields of an error response message.

    .. py:attribute:: id

        Correlation ID to match request and response.
        ``None`` is a valid ID when the error cannot be matched to a particular request.

        :type: str or int or None

    .. py:attribute:: error

        The error message. A string describing the error condition.

        :type: str

    .. py:attribute:: data

        This field may contain any JSON encodable datum that the server may want
        to return the client.

        It may contain additional information about the error condition, a partial result
        or whatever. Its presence and value are entirely optional.

        :type: Any type that can be serialized by the protocol.

    .. py:attribute:: _json_rpc_error

        The numeric error code.

        The value is usually predefined by one of the JSON protocol exceptions.
        It can be set by the developer when defining application specific exceptions.
        See :py:class:`FixedErrorMessageMixin` for an example on how to do this.

        Note that the value of this field *must* comply with the defined values in
        the standard_.

    .. _standard: https://www.jsonrpc.org/specification#error_object
    """

    def _to_dict(self):
        msg = {
            'jsonrpc': JSONRPCProtocol.JSON_RPC_VERSION,
            'id': self.unique_id,
            'error': {
                'message': str(self.error),
                'code': self._jsonrpc_error_code
            }
        }
        if hasattr(self, 'data'):
            msg['error']['data'] = self.data
        return msg

    def serialize(self):
        """Returns a serialization of the error.

        Converts the response into a bytes object that can be passed to and by the transport layer.

        :return: The serialized encoded error object.
        :rtype: bytes
        """
        return json_dumps(self._to_dict()).encode()


def _get_code_message_and_data(error):
    assert isinstance(error, (Exception, str))
    data = None
    if isinstance(error, Exception):
        if hasattr(error, 'jsonrpc_error_code'):
            code = error.jsonrpc_error_code
            msg = str(error)
            try:
                data = error.data
            except AttributeError:
                pass
        elif isinstance(error, InvalidRequestError):
            code = JSONRPCInvalidRequestError.jsonrpc_error_code
            msg = JSONRPCInvalidRequestError.message
        elif isinstance(error, MethodNotFoundError):
            code = JSONRPCMethodNotFoundError.jsonrpc_error_code
            msg = JSONRPCMethodNotFoundError.message
        elif isinstance(error, InvalidParamsError):
            code = JSONRPCInvalidParamsError.jsonrpc_error_code
            msg = JSONRPCInvalidParamsError.message
        else:
            # allow exception message to propagate
            code = JSONRPCServerError.jsonrpc_error_code
            if len(error.args) == 2:
                msg = str(error.args[0])
                data = error.args[1]
            else:
                msg = str(error)
    else:
        code = -32000
        msg = error

    return code, msg, data


class JSONRPCRequest(RPCRequest):
    def __init__(self):
        self.one_way = False
        """Request or Notification.

        :type: bool

        This flag indicates if the client expects to receive a reply (request: ``one_way = False``)
        or not (notification: ``one_way = True``).

        Note that according to the specification it is possible for the server to return an error response.
        For example if the request becomes unreadable and the server is not able to determine that it is
        in fact a notification an error should be returned. However, once the server had verified that the
        request is a notification no reply (not even an error) should be returned.
        """

        self.unique_id = None
        """Correlation ID used to match request and response.

        :type: int or str

        Generated by the client, the server copies it from request to corresponding response.
        """

        self.method = None
        """The name of the RPC function to be called.

        :type: str

        The :py:attr:`method` attribute uses the name of the function as it is known by the public.
        The :py:class:`~tinyrpc.dispatch.RPCDispatcher` allows the use of public aliases in the
        ``@public`` decorators.
        These are the names used in the :py:attr:`method` attribute.
        """

        self.args = []
        """The positional arguments of the method call.

        :type: list

        The contents of this list are the positional parameters for the :py:attr:`method` called.
        It is eventually called as ``method(*args)``.
        """

        self.kwargs = {}
        """The keyword arguments of the method call.

        :type: dict

        The contents of this dict are the keyword parameters for the :py:attr:`method` called.
        It is eventually called as ``method(**kwargs)``.
        """

    def error_respond(self, error):
        """Create an error response to this request.

        When processing the request produces an error condition this method can be used to
        create the error response object.

        :param error: Specifies what error occurred.
        :type error: str or Exception
        :returns: An error response object that can be serialized and sent to the client.
        :rtype: :py:class:`JSONRPCErrorResponse`
        """
        if self.unique_id is None:
            return None

        response = JSONRPCErrorResponse()
        response.unique_id = None if self.one_way else self.unique_id

        code, msg, data = _get_code_message_and_data(error)

        response.error = msg
        response._jsonrpc_error_code = code
        if data:
            response.data = data
        return response

    def respond(self, result):
        """Create a response to this request.

        When processing the request completed successfully this method can be used to
        create a response object.

        :param result: The result of the invoked method.
        :type result: Anything that can be encoded by JSON.
        :returns: A response object that can be serialized and sent to the client.
        :rtype: :py:class:`JSONRPCSuccessResponse`
        """
        if self.one_way or self.unique_id is None:
            return None

        response = JSONRPCSuccessResponse()

        response.result = result
        response.unique_id = self.unique_id

        return response

    def _to_dict(self):
        jdata = {
            'jsonrpc': JSONRPCProtocol.JSON_RPC_VERSION,
            'method': self.method,
        }
        if self.args:
            jdata['params'] = self.args
        if self.kwargs:
            jdata['params'] = self.kwargs
        if not self.one_way and hasattr(
                self, 'unique_id') and self.unique_id is not None:
            jdata['id'] = self.unique_id
        return jdata

    def serialize(self):
        """Returns a serialization of the request.

        Converts the request into a bytes object that can be sent to the server.

        :return: The serialized encoded request object.
        :rtype: bytes
        """
        return json_dumps(self._to_dict()).encode()


class JSONRPCBatchRequest(RPCBatchRequest):
    def create_batch_response(self):
        if self._expects_response():
            return JSONRPCBatchResponse()

    def _expects_response(self):
        for request in self:
            if isinstance(request, Exception):
                return True
            if not request.one_way and request.unique_id != None:
                return True

        return False

    def serialize(self):
        return json_dumps([req._to_dict() for req in self]).encode()


class JSONRPCBatchResponse(RPCBatchResponse):
    def serialize(self):
        return json_dumps([resp._to_dict()
                           for resp in self
                           if resp != None]).encode()


class JSONRPCProtocol(RPCBatchProtocol):
    """JSONRPC protocol implementation."""

    JSON_RPC_VERSION = "2.0"
    """Currently, only version 2.0 is supported."""

    _ALLOWED_REPLY_KEYS = sorted(['id', 'jsonrpc', 'error', 'result'])
    _ALLOWED_REQUEST_KEYS = sorted(['id', 'jsonrpc', 'method', 'params'])

    def __init__(self, *args, **kwargs):
        super(JSONRPCProtocol, self).__init__(*args, **kwargs)
        self._id_counter = 0

    def _get_unique_id(self):
        self._id_counter += 1
        return self._id_counter

    def request_factory(self):
        """Factory for request objects.

        Allows derived classes to use requests derived from :py:class:`JSONRPCRequest`.

        :rtype: :py:class:`JSONRPCRequest`
        """
        return JSONRPCRequest()

    def create_batch_request(self, requests=None):
        return JSONRPCBatchRequest(requests or [])

    def create_request(self, method, args=None, kwargs=None, one_way=False):
        """Creates a new :py:class:`JSONRPCRequest` object.

        Called by the client when constructing a request.
        JSON RPC allows either the ``args`` or ``kwargs`` argument to be set.

        :param str method: The method name to invoke.
        :param list args: The positional arguments to call the method with.
        :param dict kwargs: The keyword arguments to call the method with.
        :param bool one_way: The request is an update, i.e. it does not expect a reply.
        :return: A new request instance.
        :rtype: :py:class:`JSONRPCRequest`
        :raises InvalidRequestError: when ``args`` and ``kwargs`` are both defined.
        """
        if args and kwargs:
            raise InvalidRequestError(
                'Does not support args and kwargs at '
                'the same time'
            )

        request = self.request_factory()
        request.one_way = one_way

        if not one_way:
            request.unique_id = self._get_unique_id()

        request.method = method
        request.args = args
        request.kwargs = kwargs

        return request

    def parse_reply(self, data):
        """Deserializes and validates a response.

        Called by the client to reconstruct the serialized :py:class:`JSONRPCResponse`.

        :param bytes data: The data stream received by the transport layer containing the
            serialized request.
        :return: A reconstructed response.
        :rtype: :py:class:`JSONRPCSuccessResponse` or :py:class:`JSONRPCErrorResponse`
        :raises InvalidReplyError: if the response is not valid JSON or does not conform
            to the standard.
        """
        if isinstance(data, bytes):
            data = data.decode()

        try:
            rep = json.loads(data)
        except Exception as e:
            raise InvalidReplyError(e)

        for k in rep.keys():
            if not k in self._ALLOWED_REPLY_KEYS:
                raise InvalidReplyError('Key not allowed: %s' % k)

        if 'jsonrpc' not in rep:
            raise InvalidReplyError('Missing jsonrpc (version) in response.')

        if rep['jsonrpc'] != self.JSON_RPC_VERSION:
            raise InvalidReplyError('Wrong JSONRPC version')

        if 'id' not in rep:
            raise InvalidReplyError('Missing id in response')

        if ('error' in rep) and ('result' in rep):
            raise InvalidReplyError(
                'Reply must contain exactly one of result and error.'
            )

        if 'error' in rep:
            response = JSONRPCErrorResponse()
            error = rep['error']
            response.error = error["message"]
            response._jsonrpc_error_code = error["code"]
            if "data" in error:
                response.data = error["data"]
        else:
            response = JSONRPCSuccessResponse()
            response.result = rep.get('result', None)

        response.unique_id = rep['id']

        return response

    def parse_request(self, data):
        """Deserializes and validates a request.

        Called by the server to reconstruct the serialized :py:class:`JSONRPCRequest`.

        :param bytes data: The data stream received by the transport layer containing the
            serialized request.
        :return: A reconstructed request.
        :rtype: :py:class:`JSONRPCRequest`
        :raises JSONRPCParseError: if the ``data`` cannot be parsed as valid JSON.
        :raises JSONRPCInvalidRequestError: if the request does not comply with the standard.
        """
        if isinstance(data, bytes):
            data = data.decode()

        try:
            req = json.loads(data)
        except Exception as e:
            raise JSONRPCParseError()

        if isinstance(req, list):
            # batch request
            requests = JSONRPCBatchRequest()
            for subreq in req:
                try:
                    requests.append(self._parse_subrequest(subreq))
                except RPCError as e:
                    requests.append(e)
                except Exception as e:
                    requests.append(JSONRPCInvalidRequestError())

            if not requests:
                raise JSONRPCInvalidRequestError()
            return requests
        else:
            return self._parse_subrequest(req)

    def _parse_subrequest(self, req):
        if not isinstance(req, dict):
            raise JSONRPCInvalidRequestError()

        for k in req.keys():
            if not k in self._ALLOWED_REQUEST_KEYS:
                raise JSONRPCInvalidRequestError()

        if req.get('jsonrpc', None) != self.JSON_RPC_VERSION:
            raise JSONRPCInvalidRequestError()

        if not isinstance(req['method'], str):
            raise JSONRPCInvalidRequestError()

        request = self.request_factory()

        request.method = req['method']
        request.one_way = 'id' not in req
        if not request.one_way:
            request.unique_id = req['id']

        params = req.get('params', None)
        if params is not None:
            if isinstance(params, list):
                request.args = req['params']
            elif isinstance(params, dict):
                request.kwargs = req['params']
            else:
                raise JSONRPCInvalidParamsError()

        return request

    def raise_error(self, error):
        """Recreates the exception.

        Creates a :py:class:`~tinyrpc.protocols.jsonrpc.JSONRPCError` instance
        and raises it.
        This allows the error, message and data attributes of the original
        exception to propagate into the client code.

        The :py:attr:`~tinyrpc.protocols.RPCProtocol.raises_error` flag controls if the exception obejct is
        raised or returned.

        :returns: the exception object if it is not allowed to raise it.
        :raises JSONRPCError: when the exception can be raised.
            The exception object will contain ``message``, ``code`` and optionally a
            ``data`` property.
        """
        exc = JSONRPCError(error)
        if self.raises_errors:
            raise exc
        return exc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Dispatcher
==========

Given an RPC request the dispatcher will try to locate a server function that
implements the request and will call that function returning its return value
to the caller.
"""

import inspect

from .. import exc


def public(name=None):
    """Decorator. Mark a method as eligible for registration by a dispatcher.

    The dispatchers :py:func:`~tinyrpc.dispatch.RPCDispatcher.register_instance` function
    will do the actual registration of the marked method.

    The difference with :py:func:`~tinyrpc.dispatch.RPCDispatcher.public` is that this decorator does
    not register with a dispatcher, therefore binding the marked methods with a dispatcher is delayed
    until runtime.
    It also becomes possible to bind with multiple dispatchers.

    :param name: The name to register the function with.
    :type name: str or None

    Example:

    .. code-block:: python

        def class Baz(object):
            def not_exposed(self);
                # ...

            @public('do_something')
            def visible_method(arg1):
                # ...

        baz = Baz()
        dispatch = RPCDispatcher()
        dispatch.register_instance(baz, 'bazzies`)
        # Baz.visible_method is now callable via RPC as bazzies.do_something

    ``@public`` is a shortcut for ``@public()``.
    """
    if callable(name):
        f = name
        f._rpc_public_name = f.__name__
        return f

    def _(f):
        f._rpc_public_name = name or f.__name__
        return f

    return _


class RPCDispatcher(object):
    """Stores name-to-method mappings."""

    def __init__(self):
        self.method_map = {}
        self.subdispatchers = {}

    def public(self, name=None):
        """Convenient decorator.

        Allows easy registering of functions to this dispatcher. Example:

        .. code-block:: python

            dispatch = RPCDispatcher()

            @dispatch.public
            def foo(bar):
                # ...

            class Baz(object):
                def not_exposed(self):
                    # ...

                @dispatch.public(name='do_something')
                def visible_method(arg1)
                    # ...

        :param str name: Name to register callable with.
        """
        if callable(name):
            self.add_method(name)
            return name

        def _(f):
            self.add_method(f, name=name)
            return f

        return _

    def add_subdispatch(self, dispatcher, prefix=''):
        """Adds a subdispatcher, possibly in its own namespace.

        :param dispatcher: The dispatcher to add as a subdispatcher.
        :type dispatcher: RPCDispatcher
        :param str prefix: A prefix. All of the new subdispatchers methods will be
                       available as prefix + their original name.
        """
        self.subdispatchers.setdefault(prefix, []).append(dispatcher)

    def add_method(self, f, name=None):
        """Add a method to the dispatcher.

        :param f: Callable to be added.
        :type f: callable
        :param str name: Name to register it with. If ``None``, ``f.__name__`` will
                     be used.
        :raises ~tinyrpc.exc.RPCError: When the `name` is already registered.
        """
        assert callable(f), "method argument must be callable"
        # catches a few programming errors that are
        # commonly silently swallowed otherwise
        if not name:
            name = f.__name__

        if name in self.method_map:
            raise exc.RPCError('Name %s already registered')

        self.method_map[name] = f

    def get_method(self, name):
        """Retrieve a previously registered method.

        Checks if a method matching ``name`` has been registered.

        If :py:func:`get_method` cannot find a method, every subdispatcher
        with a prefix matching the method name is checked as well.

        :param str name: Function to find.
        :returns: The callable implementing the function.
        :rtype: callable
        :raises: :py:exc:`~tinyrpc.exc.MethodNotFoundError`
        """
        if name in self.method_map:
            return self.method_map[name]

        for prefix, subdispatchers in self.subdispatchers.items():
            if name.startswith(prefix):
                for sd in subdispatchers:
                    try:
                        return sd.get_method(name[len(prefix):])
                    except exc.MethodNotFoundError:
                        pass

        raise exc.MethodNotFoundError(name)

    def register_instance(self, obj, prefix=''):
        """Create new subdispatcher and register all public object methods on
        it.

        To be used in conjunction with the :py:func:`public`
        decorator (*not* :py:func:`RPCDispatcher.public`).

        :param obj: The object whose public methods should be made available.
        :type obj: object
        :param str prefix: A prefix for the new subdispatcher.
        """
        dispatch = self.__class__()
        for name, f in inspect.getmembers(
                obj, lambda f: callable(f) and hasattr(f, '_rpc_public_name')):
            dispatch.add_method(f, f._rpc_public_name)

        # add to dispatchers
        self.add_subdispatch(dispatch, prefix)

    def dispatch(self, request):
        """Fully handle request.

        The dispatch method determines which method to call, calls it and
        returns a response containing a result.

        No exceptions will be thrown, rather, every exception will be turned
        into a response using :py:func:`~tinyrpc.RPCRequest.error_respond`.

        If a method isn't found, a :py:exc:`~tinyrpc.exc.MethodNotFoundError`
        response will be returned. If any error occurs outside of the requested
        method, a :py:exc:`~tinyrpc.exc.ServerError` without any error
        information will be returend.

        If the method is found and called but throws an exception, the
        exception thrown is used as a response instead. This is the only case
        in which information from the exception is possibly propagated back to
        the client, as the exception is part of the requested method.

        :py:class:`~tinyrpc.RPCBatchRequest` instances are handled by handling
        all its children in order and collecting the results, then returning an
        :py:class:`~tinyrpc.RPCBatchResponse` with the results.

        :param request: The request containing the function to be called and its parameters.
        :type request: ~tinyrpc.protocols.RPCRequest
        :return: The result produced by calling the requested function.
        :rtype: ~tinyrpc.protocols.RPCResponse
        :raises ~exc.MethodNotFoundError: If the requested function is not published.
        :raises ~exc.ServerError: If some other error occurred.

        .. Note::

            The :py:exc:`~tinyrpc.exc.ServerError` is raised for any kind of exception not
            raised by the called function itself or :py:exc:`~tinyrpc.exc.MethodNotFoundError`.
        """
        if hasattr(request, 'create_batch_response'):
            results = [self._dispatch(req) for req in request]

            response = request.create_batch_response()
            if response is not None:
                response.extend(results)

            return response
        else:
            return self._dispatch(request)

    def _dispatch(self, request):
        try:
            method = self.get_method(request.method)
        except exc.MethodNotFoundError as e:
            return request.error_respond(e)
        except Exception:
            # unexpected error, do not let client know what happened
            return request.error_respond(exc.ServerError())

        # we found the method
        try:
            if self.validator is not None:
                self.validator(method, request.args, request.kwargs)
            result = method(*request.args, **request.kwargs)
        except Exception as e:
            # an error occurred within the method, return it
            return request.error_respond(e)

        # respond with result
        return request.respond(result)

    def validate_parameters(self, method, args, kwargs):
        """Verify that `*args` and `**kwargs` are appropriate parameters for `method`.

        :param method: A callable.
        :param args: List of positional arguments for `method`
        :param kwargs: Keyword arguments for `method`
        :raises ~tinyrpc.exc.InvalidParamsError:
            Raised when the provided arguments are not acceptable for `method`.
        """
        if hasattr(method, '__code__'):
            try:
                inspect.getcallargs(method, *args, **kwargs)
            except TypeError:
                raise exc.InvalidParamsError()

    validator = validate_parameters
    """Dispatched function parameter validatation.

    :type: callable
    
    By default this attribute is set to :py:func:`validate_parameters`.
    The value can be set to any callable implementing the same interface
    as :py:func:`validate_parameters` or to `None` to disable validation
    entirely.
    """




#!/usr/bin/env python
# -*- coding: utf-8 -*-

class RPCError(Exception):
    """Base class for all excetions thrown by :py:mod:`tinyrpc`."""

    def error_respond(self):
        raise NotImplementedError()

class BadRequestError(RPCError):
    """Base class for all errors that caused the processing of a request to
    abort before a request object could be instantiated."""

    def error_respond(self):
        """Create :py:class:`~tinyrpc.RPCErrorResponse` to respond the error.

        :return: A error responce instance or ``None``, if the protocol decides
                 to drop the error silently."""
        raise RuntimeError('Not implemented')


class BadReplyError(RPCError):
    """Base class for all errors that caused processing of a reply to abort
    before it could be turned in a response object."""


class InvalidRequestError(BadRequestError):
    """A request made was malformed (i.e. violated the specification) and could
    not be parsed."""


class InvalidReplyError(BadReplyError):
    """A reply received was malformed (i.e. violated the specification) and
    could not be parsed into a response."""


class MethodNotFoundError(RPCError):
    """The desired method was not found."""

class InvalidParamsError(RPCError):
    """The provided parameters do not match those of the desired method."""

class ServerError(RPCError):
    """An internal error in the RPC system occured."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Server definition.

Defines and implements a single-threaded, single-process, asynchronous server.
"""

import gevent

from . import RPCServer


class RPCServerGreenlets(RPCServer):
    """Asynchronous RPCServer.

    This implementation of :py:class:`~tinyrpc.server.RPCServer` uses
    :py:func:`gevent.spawn` to spawn new client handlers, resulting
    in asynchronous handling of clients using greenlets.
    """

    def _spawn(self, func, *args, **kwargs):
        """Spawn a handler function.

        Spawns the supplied ``func`` with ``*args`` and ``**kwargs``
        as a gevent greenlet.

        :param func: A callable to call.
        :param args: Arguments to ``func``.
        :param kwargs: Keyword arguments to ``func``.
        """
        gevent.spawn(func, *args, **kwargs)

    def start(self):
        '''
        Create a Greenlet with serve_forever so you can do a gevenet.joinall of 
        several RPCServerGreenlets  
        '''
        return gevent.spawn(self.serve_forever)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from collections import namedtuple

from .exc import RPCError
from .protocols import RPCErrorResponse

RPCCall = namedtuple('RPCCall', 'method args kwargs')
"""Defines the elements of a RPC call.

RPCCall is used with :py:meth:`~tinyrpc.client.RPCClient.call_all`
to provide the list of requests to be processed. Each request contains the
elements defined in this tuple.
"""

RPCCallTo = namedtuple('RPCCallTo', 'transport method args kwargs')
"""Defines the elements of a RPC call directed to multiple transports.

RPCCallTo is used with :py:meth:`~tinyrpc.client.RPCClient.call_all`
to provide the list of requests to be processed.
"""


class RPCClient(object):
    """Client for making RPC calls to connected servers.

    :param protocol: An :py:class:`~tinyrpc.RPCProtocol` instance.
    :param transport: A :py:class:`~tinyrpc.transports.ClientTransport`
                      instance.
    """

    def __init__(self, protocol, transport):
        self.protocol = protocol
        self.transport = transport

    def _send_and_handle_reply(
            self, req, one_way, transport=None, no_exception=None
    ):
        tport = self.transport if transport is None else transport

        # sends ...
        reply = tport.send_message(req.serialize())

        if one_way:
            # ... and be done
            return

        # ... or process the reply
        response = self.protocol.parse_reply(reply)

        if not no_exception and isinstance(response, RPCErrorResponse):
            if hasattr(self.protocol, 'raise_error') and callable(
                    self.protocol.raise_error):
                response = self.protocol.raise_error(response)
            else:
                raise RPCError(
                    'Error calling remote procedure: %s' % response.error
                )

        return response

    def call(self, method, args, kwargs, one_way=False):
        """Calls the requested method and returns the result.

        If an error occured, an :py:class:`~tinyrpc.exc.RPCError` instance
        is raised.

        :param method: Name of the method to call.
        :param args: Arguments to pass to the method.
        :param kwargs: Keyword arguments to pass to the method.
        :param one_way: Whether or not a reply is desired.
        """
        req = self.protocol.create_request(method, args, kwargs, one_way)

        rep = self._send_and_handle_reply(req, one_way)

        if one_way:
            return

        return rep.result

    def call_all(self, requests):
        """Calls the methods in the request in parallel.

        When the :py:mod:`gevent` module is already loaded it is assumed to be
        correctly initialized, including monkey patching if necessary.
        In that case the RPC calls defined by ``requests`` is performed in
        parallel otherwise the methods are called sequentially.

        :param requests: A listof either :py:class:`~tinyrpc.client.RPCCall` or :py:class:`~tinyrpc.client.RPCCallTo` elements.
                         When RPCCallTo is used each element defines a transport.
                         Otherwise the default transport set when RPCClient is
                         created is used.
        :return: A list with replies matching the order of the requests.
        """
        threads = []

        if 'gevent' in sys.modules:
            # assume that gevent is available and functional, make calls in parallel
            import gevent
            for r in requests:
                req = self.protocol.create_request(r.method, r.args, r.kwargs)
                tr = r.transport.transport if len(r) == 4 else None
                threads.append(
                    gevent.spawn(
                        self._send_and_handle_reply, req, False, tr, True
                    )
                )
            gevent.joinall(threads)
            return [t.value for t in threads]
        else:
            # call serially
            for r in requests:
                req = self.protocol.create_request(r.method, r.args, r.kwargs)
                tr = r.transport.transport if len(r) == 4 else None
                threads.append(
                    self._send_and_handle_reply(req, False, tr, True)
                )
            return threads

    def get_proxy(self, prefix='', one_way=False):
        """Convenience method for creating a proxy.

        :param prefix: Passed on to :py:class:`~tinyrpc.client.RPCProxy`.
        :param one_way: Passed on to :py:class:`~tinyrpc.client.RPCProxy`.
        :return: :py:class:`~tinyrpc.client.RPCProxy` instance."""
        return RPCProxy(self, prefix, one_way)

    def batch_call(self, calls):
        """Experimental, use at your own peril."""
        req = self.protocol.create_batch_request()

        for call_args in calls:
            req.append(self.protocol.create_request(*call_args))

        return self._send_and_handle_reply(req)


class RPCProxy(object):
    """Create a new remote proxy object.

    Proxies allow calling of methods through a simpler interface. See the
    documentation for an example.

    :param client: An :py:class:`~tinyrpc.client.RPCClient` instance.
    :param prefix: Prefix to prepend to every method name.
    :param one_way: Passed to every call of
                    :py:func:`~tinyrpc.client.call`.
    """

    def __init__(self, client, prefix='', one_way=False):
        self.client = client
        self.prefix = prefix
        self.one_way = one_way

    def __getattr__(self, name):
        """Returns a proxy function that, when called, will call a function
        name ``name`` on the client associated with the proxy.
        """
        proxy_func = lambda *args, **kwargs: self.client.call(
            self.prefix + name, args, kwargs, one_way=self.one_way
        )
        return proxy_func



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import queue as Queue

from werkzeug.wrappers import Response, Request

from . import ServerTransport


class WsgiServerTransport(ServerTransport):
    """WSGI transport.

    Requires :py:mod:`werkzeug`.

    Due to the nature of WSGI, this transport has a few pecularities: It must
    be run in a thread, greenlet or some other form of concurrent execution
    primitive.

    This is due to
    :py:func:`~tinyrpc.transports.wsgi.WsgiServerTransport.handle` blocking
    while waiting for a call to
    :py:func:`~tinyrpc.transports.wsgi.WsgiServerTransport.send_reply`.

    The parameter ``queue_class`` must be used to supply a proper queue class
    for the chosen concurrency mechanism (i.e. when using :py:mod:`gevent`,
    set it to :py:class:`gevent.queue.Queue`).

    :param max_content_length: The maximum request content size allowed. Should
                               be set to a sane value to prevent DoS-Attacks.
    :param queue_class: The Queue class to use.
    :param allow_origin: The ``Access-Control-Allow-Origin`` header. Defaults
                         to ``*`` (so change it if you need actual security).
    """
    def __init__(self, max_content_length=4096, queue_class=Queue.Queue,
                       allow_origin='*'):
        self._queue_class = queue_class
        self.messages = queue_class()
        self.max_content_length = max_content_length
        self.allow_origin = allow_origin

    def receive_message(self):
        return self.messages.get()

    def send_reply(self, context, reply):
        context.put(reply)

    def handle(self, environ, start_response):
        """WSGI handler function.

        The transport will serve a request by reading the message and putting
        it into an internal buffer. It will then block until another
        concurrently running function sends a reply using :py:meth:`send_reply`.

        The reply will then be sent to the client being handled and handle will
        return.
        """
        request = Request(environ)
        request.max_content_length = self.max_content_length

        access_control_headers = {
            'Access-Control-Allow-Methods': 'POST',
            'Access-Control-Allow-Origin': self.allow_origin,
            'Access-Control-Allow-Headers': \
                'Content-Type, X-Requested-With, Accept, Origin'
        }

        if request.method == 'OPTIONS':
            response = Response(headers=access_control_headers)

        elif request.method == 'POST':
            # message is encoded in POST, read it...
            msg = request.stream.read()

            # create new context
            context = self._queue_class()

            self.messages.put((context, msg))

            # ...and send the reply
            response = Response(context.get(), headers=access_control_headers)
        else:
            # nothing else supported at the moment
            response = Response('Only POST supported', 405)

        return response(environ, start_response)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class ServerTransport(object):
    """Abstract base class for all server transports.

    The server side implementation of the transport component.
    Requests and replies encoded by the protocol component are
    exchanged between client and server using the :py:class:`ServerTransport`
    and :py:class:`ClientTransport` classes.
    """

    def receive_message(self):
        """Receive a message from the transport.

        Blocks until a message has been received.
        May return an opaque context object to its caller that should be passed on to
        :py:func:`~tinyrpc.transport.ServerTransport.send_reply` to identify
        the transport or requester later on.
        Use and function of the context object are entirely controlled by the transport
        instance.

        The message must be treated as a binary entity as only the protocol
        level will know how to interpret the message.

        If the transport encodes the message in some way, the opposite end
        is responsible for decoding it before it is passed to either client
        or server.

        :return: A tuple consisting of ``(context, message)``.
            Where ``context`` can be any valid Python type and
            ``message`` must be a :py:class:`bytes` object.
        """
        raise NotImplementedError()

    def send_reply(self, context, reply):
        """Sends a reply to a client.

        The client is usually identified by passing ``context`` as returned
        from the original :py:meth:`receive_message` call.

        The reply must be a bytes object since only the protocol
        level will know how to construct the reply.

        :param any context: A context returned by :py:meth:`receive_message`.
        :param bytes reply: The reply to return to the client.
        """
        raise NotImplementedError


class ClientTransport(object):
    """Abstract base class for all client transports.

    The client side implementation of the transport component.
    Requests and replies encoded by the protocol component are
    exchanged between client and server using the :py:class:`ServerTransport`
    and :py:class:`ClientTransport` classes.    """

    def send_message(self, message, expect_reply=True):
        """Send a message to the server and possibly receive a reply.

        Sends a message to the connected server.

        The message must be treated as a binary entity as only the protocol
        level will know how to interpret the message.

        If the transport encodes the message in some way, the opposite end
        is responsible for decoding it before it is passed to either client
        or server.

        This function will block until the reply has been received.

        :param bytes message: The request to send to the server.
        :param bool expect_reply: Some protocols allow notifications for which a
            reply is not expected. When this flag is ``False`` the transport may
            not wait for a response from the server.
            **Note** that it is still the responsibility of the transport layer how
            to implement this. It is still possible that the server sends some form
            of reply regardless the value of this flag.
        :return: The servers reply to the request.
        :rtype: bytes
        """
        raise NotImplementedError
